module Dag
  class Job < Model
    def initialize(api, job_info)
      super(api)
      update_parameters(job_info)
    end

    attr_reader :id, :status, :dsl, :process_engine, :start_at, :type, :cluster, :cluster_rebooted, :label, :stage, :progress,
      :access_key_id, :query, :output_format, :output_resource_path, :job_id, :schema, :input_object_keys,
      :input_format, :output_database, :output_table

    def finished?
      @status == 'finished'
    end

    def running?
      @status == 'running'
    end

    def split?
      @type == 'split'
    end

    def hive?
      @type == 'select'
    end

    def cluster_rebooted?
      !!@cluster_rebooted
    end

    def reload
      job_info = @api.query_info(@id)
      update_parameters(job_info)
    end

    def kill
      validate_cancel_condition
      @api.query_cancel(@id)
    end

    alias :cancel :kill

    def download_urls(time_limit = 30)
      raise Dag::Client::StatusInvalid.new("job status is not finished") unless finished?
      expire_at = time_limit.minutes.since.to_i
      object_uri = URI.parse(@output_resource_path)
      bucket = object_uri.host
      object_path = object_uri.path[1..-1]
      object_path += '/' unless object_path.end_with? '/'
      bucket_objects = @api.objects(bucket, prefix: object_path).objects
      bucket_objects.map do |object|
        path = if @api.force_path_style?
                 "/#{bucket}/#{object}"
               else
                 "/#{object}"
               end

        parameters = {
          "Expires" => expire_at,
          "IIJGIOAccessKeyId" => @api.apikey,
          "Signature" => @api.download_signature(expire_at, bucket, path)
        }

        uri = URI.parse(@api.storage_api)
        url = if @api.force_path_style?
                "http://#{uri.host}"
              else
                "http://#{bucket}.#{uri.host}"
              end
        url += ":#{uri.port}" unless uri.port == 80 || uri.port == 443

        File.join(url, "#{path}?#{parameters.to_param}")
      end
    end

    def log
      validate_log_condition
      log_info = @api.query_log(@id)
      log_info ? log_info['log'] : ''
    end

    def validate_log_condition
      if split?
        raise Dag::Client::JobTypeInvalid.new("job type is not select")
      end

      if cluster_rebooted?
        raise Dag::Client::ClusterRebooted.new("cluster is rebooted")
      end
    end

    def validate_cancel_condition
      unless running?
        raise Dag::Client::StatusInvalid.new("job status is not running")
      end
    end

    private

    def update_parameters(job)
      @id = job['id']
      @status = job['status']
      @process_engine = job['processEngine']
      @dsl = job['dsl']
      @cluster = job['clusterName']
      @cluster_rebooted = job['clusterRebooted']
      @start_at = Time.parse(job['startTime']) if job['startTime']
      @access_key_id = job['accessKeyId']
      @query = job['query']
      @output_format = job['outputFormat']
      @output_resource_path = job['outputResourcePath']
      @type = job['type']
      @label = job['label']
      @stage = job['stage']
      @progress = job['progress']
      @job_id = job['jobId']
      @schema = job['schema']
      @input_object_keys = job['inputObjectKeys']
      @input_format = job['inputFormat']
      @output_database = job['outputDatabase']
      @output_table = job['outputTable']
    end
  end
end
